#include "bno08x.h"
#include "INTRINS.H"
#include "math.h"

void delay(unsigned int t) //@24.000MHz
{
    unsigned long edata i;
    for (; t > 0; t--)
    {
        _nop_();
        _nop_();
        _nop_();
        i = 5998UL;
        while (i)
            i--;
    }
}

// Global Variables ȫֱ
unsigned char shtpHeader[4]; // Each packet has a header of 4 chars ÿһ4ַıͷ
unsigned char shtpData[MAX_PACKET_SIZE];
unsigned char sequenceNumber[6] = {0, 0, 0, 0, 0, 0}; // There are 6 com channels. Each channel has its own seqnum 6comƵÿƵԼ
unsigned char commandSequenceNumber = 0;              // Commands have a seqNum as well. These are inside command packet, the header uses its own seqNum per channel ҲһкšЩкݰڲͷΪÿͨʹкš
unsigned long metaData[MAX_METADATA_SIZE];                 // There is more than 10 words in a metadata record but we'll stop at Q point 3 Ԫݼ¼г10ʣǽQ3ֹͣ
// Variables 
unsigned char _deviceAddress;// I2C ַsetI2CAddress һ㡣


// These are the raw sensor values pulled from the user requested Input Report ЩǴû뱨ȡԭʼֵ
unsigned int rawAccelX, rawAccelY, rawAccelZ, accelAccuracy;
unsigned int rawLinAccelX, rawLinAccelY, rawLinAccelZ, accelLinAccuracy;
unsigned int rawGyroX, rawGyroY, rawGyroZ, gyroAccuracy;
unsigned int rawMagX, rawMagY, rawMagZ, magAccuracy;
unsigned int rawQuatI, rawQuatJ, rawQuatK, rawQuatReal, rawQuatRadianAccuracy, quatAccuracy;
unsigned int stepCount;
unsigned char stabilityClassifier;
unsigned char activityClassifier;
unsigned char *_activityConfidences; // Array that store the confidences of the 9 possible activities 飬洢9ܻŶ

// These Q values are defined in the datasheet but can also be obtained by querying the meta data records ЩQֵݱж壬ҲͨѯԪݼ¼
// See the read metadata example for more info йظϢĶȡԪʾ
int rotationVector_Q1 = 14;
int accelerometer_Q1 = 8;
int linear_accelerometer_Q1 = 8;
int gyro_Q1 = 9;
int magnetometer_Q1 = 4;
//
char dataAvailable(void)
{
    if (receivePacket() == 1)
    {
        // Check to see if this packet is a sensor reporting its data to us ݰǷǱݵĴ
        if (shtpHeader[2] == CHANNEL_REPORTS && shtpData[0] == SHTP_REPORT_BASE_TIMESTAMP)
        {
            parseInputReport(); // This will update the rawAccelX, etc variables depending on which feature report is found ⽫ҵԱrawAccelXȱ
            return (1);
        }
    }
    return (0);
}

// This function pulls the data from the input report ú뱨ȡ
// The input reports vary in length so this function stores the various 16-bit values as globals 뱨ĳȸͬ˸ú16λֵ洢Ϊȫֵ

// Unit responds with packet that contains the following: UnitðݵݰӦ
// shtpHeader[0:3]: First, a 4 byte header shtpHeader[0:3]ȣһ4ֽڵıͷ
// shtpData[0:4]: Then a 5 byte timestamp of microsecond clicks since reading was taken shtpData[0:4]Ȼһ5ֽڵʱ΢ΪλӶȡʼ
// shtpData[5 + 0]: Then a feature report ID (0x01 for Accel, 0x05 for Rotation Vector) shtpData[5 + 0]ȻһID 0x01 for Accel, 0x05 for Rotation Vector
// shtpData[5 + 1]: Sequence number (See 6.5.18.2) shtpData[5 + 1]кţμ6.5.18.2
// shtpData[5 + 2]: Status shtpData[5 + 2]״̬
// shtpData[3]: Delay shtpData[3]:ӳ
// shtpData[4:5]: i/accel x/gyro x/etc shtpData[4:5]: i/ٶ x/ x/
// shtpData[6:7]: j/accel y/gyro y/etc shtpData[6:7]: j/ٶ y/ y/
// shtpData[8:9]: k/accel z/gyro z/etc shtpData[8:9]: k/ٶ z/ z/
// shtpData[10:11]: real/gyro temp/etc
// shtpData[12:13]: Accuracy estimate  shtpData[12:13]ȹ
void parseInputReport(void)
{
    // Calculate the number of data bytes in this packet ݰеֽ
    int dataLength = ((unsigned int)shtpHeader[1] << 8 | shtpHeader[0]);

    unsigned char status = shtpData[5 + 2] & 0x03; // Get status bits ȡ״̬λ
    unsigned int data1 = (unsigned int)shtpData[5 + 5] << 8 | shtpData[5 + 4];
    unsigned int data2 = (unsigned int)shtpData[5 + 7] << 8 | shtpData[5 + 6];
    unsigned int data3 = (unsigned int)shtpData[5 + 9] << 8 | shtpData[5 + 8];
    unsigned int data4 = 0;
    unsigned int data5 = 0;
    unsigned char x = 0;
    dataLength &= ~(1 << 15); // Clear the MSbit. This bit indicates if this package is a continuation of the last. MSbitλʾǷһ
    // Ignore it for now. TODO catch this as an error and exit ʱTODO˲Ϊ˳

    dataLength -= 4; // Remove the header bytes from the data count ݼɾͷֽ

    if (dataLength - 5 > 9)
    {
        data4 = (unsigned int)shtpData[5 + 11] << 8 | shtpData[5 + 10];
    }
    if (dataLength - 5 > 11)
    {
        data5 = (unsigned int)shtpData[5 + 13] << 8 | shtpData[5 + 12];
    }

    // Store these generic values to their proper global variable Щֵ洢ʵȫֱ
    if (shtpData[5] == SENSOR_REPORTID_ACCELEROMETER)
    {
        accelAccuracy = status;
        rawAccelX = data1;
        rawAccelY = data2;
        rawAccelZ = data3;
    }
    else if (shtpData[5] == SENSOR_REPORTID_LINEAR_ACCELERATION)
    {
        accelLinAccuracy = status;
        rawLinAccelX = data1;
        rawLinAccelY = data2;
        rawLinAccelZ = data3;
    }
    else if (shtpData[5] == SENSOR_REPORTID_GYROSCOPE)
    {
        gyroAccuracy = status;
        rawGyroX = data1;
        rawGyroY = data2;
        rawGyroZ = data3;
    }
    else if (shtpData[5] == SENSOR_REPORTID_MAGNETIC_FIELD)
    {
        magAccuracy = status;
        rawMagX = data1;
        rawMagY = data2;
        rawMagZ = data3;
    }
    else if (shtpData[5] == SENSOR_REPORTID_ROTATION_VECTOR || shtpData[5] == SENSOR_REPORTID_GAME_ROTATION_VECTOR)
    {
        quatAccuracy = status;
        rawQuatI = data1;
        rawQuatJ = data2;
        rawQuatK = data3;
        rawQuatReal = data4;
        rawQuatRadianAccuracy = data5; // Only available on rotation vector, not game rot vector
    }
    else if (shtpData[5] == SENSOR_REPORTID_STEP_COUNTER)
    {
        stepCount = data3; // Bytes 8/9
    }
    else if (shtpData[5] == SENSOR_REPORTID_STABILITY_CLASSIFIER)
    {
        stabilityClassifier = shtpData[5 + 4]; // Byte 4 only
    }
    else if (shtpData[5] == SENSOR_REPORTID_PERSONAL_ACTIVITY_CLASSIFIER)
    {
        activityClassifier = shtpData[5 + 5]; // Most likely state

        // Load activity classification confidences into the array
        for (x = 0; x < 9; x++)                            // Hardcoded to max of 9. TODO - bring in array size
            _activityConfidences[x] = shtpData[5 + 6 + x]; // 5 bytes of timestamp, byte 6 is first confidence byte
    }
    else
    {
        // This sensor report ID is unhandled.
        // See reference manual to add additional feature reports as needed
    }

    // TODO additional feature reports may be strung together. Parse them all.
}

// Return the rotation vector quaternion I
float getQuatI()
{
    float quat = qToFloat(rawQuatI, (unsigned char)rotationVector_Q1);
    return (quat);
}

// Return the rotation vector quaternion J
float getQuatJ()
{
    float quat = qToFloat(rawQuatJ, (unsigned char)rotationVector_Q1);
    return (quat);
}

// Return the rotation vector quaternion K
float getQuatK()
{
    float quat = qToFloat(rawQuatK, (unsigned char)rotationVector_Q1);
    return (quat);
}

// Return the rotation vector quaternion Real
float getQuatReal()
{
    float quat = qToFloat(rawQuatReal, (unsigned char)rotationVector_Q1);
    return (quat);
}

// Return the rotation vector accuracy
float getQuatRadianAccuracy()
{
    float quat = qToFloat(rawQuatRadianAccuracy, (unsigned char)rotationVector_Q1);
    return (quat);
}

// Return the acceleration component
unsigned char getQuatAccuracy()
{
    return (quatAccuracy);
}

// Return the acceleration component
float getAccelX()
{
    float accel = qToFloat(rawAccelX, (unsigned char)accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
float getAccelY()
{
    float accel = qToFloat(rawAccelY, (unsigned char)accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
float getAccelZ()
{
    float accel = qToFloat(rawAccelZ, (unsigned char)accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
unsigned char getAccelAccuracy()
{
    return (accelAccuracy);
}

// linear acceleration, i.e. minus gravity

// Return the acceleration component
float getLinAccelX()
{
    float accel = qToFloat(rawLinAccelX, (unsigned char)linear_accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
float getLinAccelY()
{
    float accel = qToFloat(rawLinAccelY, (unsigned char)linear_accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
float getLinAccelZ()
{
    float accel = qToFloat(rawLinAccelZ, (unsigned char)linear_accelerometer_Q1);
    return (accel);
}

// Return the acceleration component
unsigned char getLinAccelAccuracy()
{
    return (accelLinAccuracy);
}

// Return the gyro component
float getGyroX()
{
    float gyro = qToFloat(rawGyroX, (unsigned char)gyro_Q1);
    return (gyro);
}

// Return the gyro component
float getGyroY()
{
    float gyro = qToFloat(rawGyroY, (unsigned char)gyro_Q1);
    return (gyro);
}

// Return the gyro component
float getGyroZ()
{
    float gyro = qToFloat(rawGyroZ, (unsigned char)gyro_Q1);
    return (gyro);
}

// Return the gyro component
unsigned char getGyroAccuracy()
{
    return (gyroAccuracy);
}

// Return the magnetometer component
float getMagX()
{
    float mag = qToFloat(rawMagX, (unsigned char)magnetometer_Q1);
    return (mag);
}

// Return the magnetometer component
float getMagY()
{
    float mag = qToFloat(rawMagY, (unsigned char)magnetometer_Q1);
    return (mag);
}

// Return the magnetometer component
float getMagZ()
{
    float mag = qToFloat(rawMagZ, (unsigned char)magnetometer_Q1);
    return (mag);
}

// Return the mag component
unsigned char getMagAccuracy()
{
    return (magAccuracy);
}

// Return the step count
unsigned int getStepCount()
{
    return (stepCount);
}

// Return the stability classifier
unsigned char getStabilityClassifier()
{
    return (stabilityClassifier);
}

// Return the activity classifier
unsigned char getActivityClassifier()
{
    return (activityClassifier);
}

// Given a record ID, read the Q1 value from the metaData record in the FRS (ya, it's complicated)
// Q1 is used for all sensor data calculations
int getQ1(unsigned int recordID)
{
    // Q1 is always the lower 16 bits of word 7
    unsigned int q = readFRSword(recordID, 7) & 0xFFFF; // Get word 7, lower 16 bits
    return (q);
}

// Given a record ID, read the Q2 value from the metaData record in the FRS
// Q2 is used in sensor bias
int getQ2(unsigned int recordID)
{
    // Q2 is always the upper 16 bits of word 7
    unsigned int q = readFRSword(recordID, 7) >> 16; // Get word 7, upper 16 bits
    return (q);
}

// Given a record ID, read the Q3 value from the metaData record in the FRS
// Q3 is used in sensor change sensitivity
int getQ3(unsigned int recordID)
{
    // Q3 is always the upper 16 bits of word 8
    unsigned int q = readFRSword(recordID, 8) >> 16; // Get word 8, upper 16 bits
    return (q);
}

// Given a record ID, read the resolution value from the metaData record in the FRS for a given sensor
float getResolution(unsigned int recordID)
{
    // The resolution Q value are 'the same as those used in the sensor's input report'
    // This should be Q1.
    int Q = getQ1(recordID);

    // Resolution is always word 2
    unsigned long value = readFRSword(recordID, 2); // Get word 2

    float resolution = qToFloat((unsigned int)value, (unsigned char)Q);

    return (resolution);
}

// Given a record ID, read the range value from the metaData record in the FRS for a given sensor
float getRange(unsigned int recordID)
{
    // The resolution Q value are 'the same as those used in the sensor's input report'
    // This should be Q1.
    int Q = getQ1(recordID);

    // Range is always word 1
    unsigned long value = readFRSword(recordID, 1); // Get word 1

    float range = qToFloat((unsigned int)value, (unsigned char)Q);

    return (range);
}

// Given a record ID and a word number, look up the word data
// Helpful for pulling out a Q value, range, etc.
// Use readFRSdata for pulling out multi-word objects for a sensor (Vendor data for example)
unsigned long readFRSword(unsigned int recordID, unsigned char wordNumber)
{
    if (readFRSdata(recordID, wordNumber, 1) == 1) // Get word number, just one word in length from FRS
        return (metaData[0]);                      // Return this one word

    return (0); // Error
}

// Ask the sensor for data from the Flash Record System
// See 6.3.6 page 40, FRS Read Request
void frsReadRequest(unsigned int recordID, unsigned int readOffset, unsigned int blockSize)
{
    shtpData[0] = SHTP_REPORT_FRS_READ_REQUEST; // FRS Read Request
    shtpData[1] = 0;                            // Reserved
    shtpData[2] = (readOffset >> 0) & 0xFF;     // Read Offset LSB
    shtpData[3] = (readOffset >> 8) & 0xFF;     // Read Offset MSB
    shtpData[4] = (recordID >> 0) & 0xFF;       // FRS Type LSB
    shtpData[5] = (recordID >> 8) & 0xFF;       // FRS Type MSB
    shtpData[6] = (blockSize >> 0) & 0xFF;      // Block size LSB
    shtpData[7] = (blockSize >> 8) & 0xFF;      // Block size MSB

    // Transmit packet on channel 2, 8 bytes
    sendPacket(CHANNEL_CONTROL, 8);
}

// Given a sensor or record ID, and a given start/stop bytes, read the data from the Flash Record System (FRS) for this sensor
// Returns 1 if metaData array is loaded successfully
// Returns 0 if failure
char readFRSdata(unsigned int recordID, unsigned char startLocation, unsigned char wordsToRead)
{
    unsigned char spot = 0;
    unsigned char dataLength = 0;
    unsigned char frsStatus = 0;
    unsigned long data0 = 0;
    unsigned long data1 = 0;
    // First we send a Flash Record System (FRS) request
    frsReadRequest(recordID, startLocation, wordsToRead); // From startLocation of record, read a # of words

    // Read bytes until FRS reports that the read is complete
    while (1)
    {
        // Now we wait for response
        while (1)
        {
            unsigned char counter = 0;
            while (receivePacket() == 0)
            {
                if (counter++ > 100)
                    return (0); // Give up
                delay(1);
            }

            // We have the packet, inspect it for the right contents
            // See page 40. Report ID should be 0xF3 and the FRS types should match the thing we requested
            if (shtpData[0] == SHTP_REPORT_FRS_READ_RESPONSE)
                if (((unsigned int)shtpData[13] << 8 | shtpData[12]) == recordID)
                    break; // This packet is one we are looking for
        }

        dataLength = shtpData[1] >> 4;
        frsStatus = shtpData[1] & 0x0F;

        data0 = (unsigned long)shtpData[7] << 24 | (unsigned long)shtpData[6] << 16 | (unsigned long)shtpData[5] << 8 | (unsigned long)shtpData[4];
        data1 = (unsigned long)shtpData[11] << 24 | (unsigned long)shtpData[10] << 16 | (unsigned long)shtpData[9] << 8 | (unsigned long)shtpData[8];

        // Record these words to the metaData array
        if (dataLength > 0)
        {
            metaData[spot++] = data0;
        }
        if (dataLength > 1)
        {
            metaData[spot++] = data1;
        }

        if (frsStatus == 3 || frsStatus == 6 || frsStatus == 7)
        {
            return (1); // FRS status is read completed! We're done!
        }
    }
}

// Send command to reset IC
// Read all advertisement packets from sensor
// The sensor has been seen to reset twice if we attempt too much too quickly.
// This seems to work reliably.
void softReset(void)
{
    shtpData[0] = 1; // Reset

    // Attempt to start communication with sensor
    sendPacket(CHANNEL_EXECUTABLE, 1); // Transmit packet on channel 1, 1 byte

    // Read all incoming data and flush it
    delay(50);
    while (receivePacket() == 1)
        ;
    delay(50);
    while (receivePacket() == 1)
        ;
}

// Get the reason for the last reset
// 1 = POR, 2 = Internal reset, 3 = Watchdog, 4 = External reset, 5 = Other
unsigned char resetReason()
{
    shtpData[0] = SHTP_REPORT_PRODUCT_ID_REQUEST; // Request the product ID and reset info
    shtpData[1] = 0;                              // Reserved

    // Transmit packet on channel 2, 2 bytes
    sendPacket(CHANNEL_CONTROL, 2);

    // Now we wait for response
    if (receivePacket() == 1)
    {
        if (shtpData[0] == SHTP_REPORT_PRODUCT_ID_RESPONSE)
        {
            return (shtpData[1]);
        }
    }

    return (0);
}

// Given a register value and a Q point, convert to float
// See https://en.wikipedia.org/wiki/Q_(number_format)
float qToFloat(int fixedPointValue, unsigned char qPoint)
{
    float qFloat = fixedPointValue;
    qFloat *= pow(2, qPoint * -1);
    return (qFloat);
}

// Sends the packet to enable the rotation vector
void enableRotationVector(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_ROTATION_VECTOR, timeBetweenReports, 0);
}

// Sends the packet to enable the rotation vector
void enableGameRotationVector(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_GAME_ROTATION_VECTOR, timeBetweenReports, 0);
}

// Sends the packet to enable the accelerometer
void enableAccelerometer(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_ACCELEROMETER, timeBetweenReports, 0);
}

// Sends the packet to enable the accelerometer
void enableLinearAccelerometer(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_LINEAR_ACCELERATION, timeBetweenReports, 0);
}

// Sends the packet to enable the gyro
void enableGyro(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_GYROSCOPE, timeBetweenReports, 0);
}

// Sends the packet to enable the magnetometer
void enableMagnetometer(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_MAGNETIC_FIELD, timeBetweenReports, 0);
}

// Sends the packet to enable the step counter
void enableStepCounter(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_STEP_COUNTER, timeBetweenReports, 0);
}

// Sends the packet to enable the Stability Classifier
void enableStabilityClassifier(unsigned int timeBetweenReports)
{
    setFeatureCommand(SENSOR_REPORTID_STABILITY_CLASSIFIER, timeBetweenReports, 0);
}

// Sends the commands to begin calibration of the accelerometer
void calibrateAccelerometer()
{
    sendCalibrateCommand(CALIBRATE_ACCEL);
}

// Sends the commands to begin calibration of the gyro
void calibrateGyro()
{
    sendCalibrateCommand(CALIBRATE_GYRO);
}

// Sends the commands to begin calibration of the magnetometer
void calibrateMagnetometer()
{
    sendCalibrateCommand(CALIBRATE_MAG);
}

// Sends the commands to begin calibration of the planar accelerometer
void calibratePlanarAccelerometer()
{
    sendCalibrateCommand(CALIBRATE_PLANAR_ACCEL);
}

// See 2.2 of the Calibration Procedure document 1000-4044
void calibrateAll()
{
    sendCalibrateCommand(CALIBRATE_ACCEL_GYRO_MAG);
}

void endCalibration()
{
    sendCalibrateCommand(CALIBRATE_STOP); // Disables all calibrations
}

// Given a sensor's report ID, this tells the BNO080 to begin reporting the values
// Also sets the specific config word. Useful for personal activity classifier
void setFeatureCommand(unsigned char reportID, unsigned int timeBetweenReports, unsigned long specificConfig)
{
    long microsBetweenReports = (long)timeBetweenReports * 1000L;

    shtpData[0] = SHTP_REPORT_SET_FEATURE_COMMAND;     // Set feature command. Reference page 55
    shtpData[1] = reportID;                            // Feature Report ID. 0x01 = Accelerometer, 0x05 = Rotation vector
    shtpData[2] = 0;                                   // Feature flags
    shtpData[3] = 0;                                   // Change sensitivity (LSB)
    shtpData[4] = 0;                                   // Change sensitivity (MSB)
    shtpData[5] = (microsBetweenReports >> 0) & 0xFF;  // Report interval (LSB) in microseconds. 0x7A120 = 500ms
    shtpData[6] = (microsBetweenReports >> 8) & 0xFF;  // Report interval
    shtpData[7] = (microsBetweenReports >> 16) & 0xFF; // Report interval
    shtpData[8] = (microsBetweenReports >> 24) & 0xFF; // Report interval (MSB)
    shtpData[9] = 0;                                   // Batch Interval (LSB)
    shtpData[10] = 0;                                  // Batch Interval
    shtpData[11] = 0;                                  // Batch Interval
    shtpData[12] = 0;                                  // Batch Interval (MSB)
    shtpData[13] = (specificConfig >> 0) & 0xFF;       // Sensor-specific config (LSB)
    shtpData[14] = (specificConfig >> 8) & 0xFF;       // Sensor-specific config
    shtpData[15] = (specificConfig >> 16) & 0xFF;      // Sensor-specific config
    shtpData[16] = (specificConfig >> 24) & 0xFF;      // Sensor-specific config (MSB)

    // Transmit packet on channel 2, 17 bytes
    sendPacket(CHANNEL_CONTROL, 17);
}

// Tell the sensor to do a command
// See 6.3.8 page 41, Command request
// The caller is expected to set P0 through P8 prior to calling
void sendCommand(unsigned char command)
{
    shtpData[0] = SHTP_REPORT_COMMAND_REQUEST; // Command Request
    shtpData[1] = commandSequenceNumber++;     // Increments automatically each function call
    shtpData[2] = command;                     // Command

    // Caller must set these
    /*shtpData[3] = 0; //P0
    shtpData[4] = 0; //P1
    shtpData[5] = 0; //P2
    shtpData[6] = 0;
    shtpData[7] = 0;
    shtpData[8] = 0;
    shtpData[9] = 0;
    shtpData[10] = 0;
    shtpData[11] = 0;*/

    // Transmit packet on channel 2, 12 bytes
    sendPacket(CHANNEL_CONTROL, 12);
}

// This tells the BNO080 to begin calibrating
// See page 50 of reference manual and the 1000-4044 calibration doc
void sendCalibrateCommand(unsigned char thingToCalibrate)
{
    unsigned char x = 0;
    /*shtpData[3] = 0; //P0 - Accel Cal Enable
    shtpData[4] = 0; //P1 - Gyro Cal Enable
    shtpData[5] = 0; //P2 - Mag Cal Enable
    shtpData[6] = 0; //P3 - Subcommand 0x00
    shtpData[7] = 0; //P4 - Planar Accel Cal Enable
    shtpData[8] = 0; //P5 - Reserved
    shtpData[9] = 0; //P6 - Reserved
    shtpData[10] = 0; //P7 - Reserved
    shtpData[11] = 0; //P8 - Reserved*/

    for (x = 3; x < 12; x++) // Clear this section of the shtpData array
        shtpData[x] = 0;

    if (thingToCalibrate == CALIBRATE_ACCEL)
        shtpData[3] = 1;
    else if (thingToCalibrate == CALIBRATE_GYRO)
        shtpData[4] = 1;
    else if (thingToCalibrate == CALIBRATE_MAG)
        shtpData[5] = 1;
    else if (thingToCalibrate == CALIBRATE_PLANAR_ACCEL)
        shtpData[7] = 1;
    else if (thingToCalibrate == CALIBRATE_ACCEL_GYRO_MAG)
    {
        shtpData[3] = 1;
        shtpData[4] = 1;
        shtpData[5] = 1;
    }
    else if (thingToCalibrate == CALIBRATE_STOP)
        ; // Do nothing, bytes are set to zero

    // Using this shtpData packet, send a command
    sendCommand(COMMAND_ME_CALIBRATE);
}

// This tells the BNO080 to save the Dynamic Calibration Data (DCD) to flash
// See page 49 of reference manual and the 1000-4044 calibration doc
void saveCalibration()
{
    unsigned char x = 0;
    /*shtpData[3] = 0; //P0 - Reserved
    shtpData[4] = 0; //P1 - Reserved
    shtpData[5] = 0; //P2 - Reserved
    shtpData[6] = 0; //P3 - Reserved
    shtpData[7] = 0; //P4 - Reserved
    shtpData[8] = 0; //P5 - Reserved
    shtpData[9] = 0; //P6 - Reserved
    shtpData[10] = 0; //P7 - Reserved
    shtpData[11] = 0; //P8 - Reserved*/

    for (x = 3; x < 12; x++) // Clear this section of the shtpData array
        shtpData[x] = 0;

    // Using this shtpData packet, send a command
    sendCommand(COMMAND_DCD); // Save DCD command
}

typedef unsigned char u8;
typedef unsigned int u16;
typedef unsigned long u32;
sbit SCL = P2^5;
sbit SDA = P2^4;

void Delay_us(unsigned int us) 
{
    while (us--) 
	{
		unsigned long edata i;

		_nop_();
		_nop_();
		_nop_();
		i = 4UL;
		while (i) i--;
    }
}
void I2C_Init(void)
{
    SCL = 1;
    SDA = 1;
    Delay_us(10);
}



//void Wait()
//{
//	EAXFR = 1; // չĴ(XFR)ʹ
//	while (!(I2CMSST & 0x40))
//       ;
//    I2CMSST &= ~0x40;
//}

void Start()
{
    SDA = 1;
    SCL = 1;
    Delay_us(10);
    SDA = 0;
    Delay_us(10);
    SCL = 0;
}

void SendData(char dat)
{
    unsigned char i;

    for (i = 0; i < 8; i++) 
	{
        SDA = (dat & 0x80) ? 1 : 0;  // λ
        dat <<= 1;
        SCL = 1;
        Delay_us(10);
        SCL = 0;
        Delay_us(10);
    }

    // ȡACK
    SDA = 1;
    SCL = 1;
    Delay_us(10);
    SCL = 0;
}

void RecvACK()
{
    // ȡACKźţѰSendDataУʡԻպ
    SDA = 1;
    SCL = 1;
    Delay_us(10);
    SCL = 0;
}

char RecvData()
{
   unsigned char i, dat = 0;

    SDA = 1;  // ͷSDAⲿ

    for (i = 0; i < 8; i++) 
	{
        SCL = 1;
        Delay_us(10);
        dat <<= 1;
        if (SDA) dat |= 0x01;
        SCL = 0;
        Delay_us(10);
    }

    return dat;
}

void SendACK()
{
    SDA = 0;
    SCL = 1;
    Delay_us(10);
    SCL = 0;
    Delay_us(10);
}

void SendNAK()
{
    SDA = 1;
    SCL = 1;
    Delay_us(10);
    SCL = 0;
    Delay_us(10);
}

void Stop()
{
    SDA = 0;
    SCL = 1;
    Delay_us(10);
    SDA = 1;
    Delay_us(10);
}

#define SLAW 0x96
#define SLAR 0x97

void WriteNbyte(u8 *p, u8 number) /*  WordAddress,First Data Address,Byte lenth   */
{
    Start();        // ʼź
    SendData(SLAW); // 豸ַ + д7λַ1λλΪ0
    RecvACK();      // ACKӦ

    while (number--)
    {
        SendData(*p++); // 
        RecvACK();      // ACKӦ
    }

    Stop(); // ֹͣź
}

void ReadNbyte(u8 *p, u8 number) /*  WordAddress,First Data Address,Byte lenth   */
{
    Start();        // ʼź
    SendData(SLAR); // 豸ַ + 7λַ1λλΪ1
    RecvACK();      // ACKӦ

    while (number--)
    {
        *p++ = RecvData(); // 
        if (number > 0)
            SendACK(); // ݣACK
        else
            SendNAK(); // һֽڷNACK
    }

    Stop(); // ֹͣź
}

u8 i2c_tmp[32] = {0};

// Ƿκοõ
// ݰݶ shtpData 
char receivePacket(void)
{
   int dataLength = 0;
    unsigned char packetLSB = 0;
    unsigned char packetMSB = 0;
    unsigned char channelNumber = 0;
    unsigned char sequenceNumber = 0;

    // ȡͷ4ֽڣ
    ReadNbyte(i2c_tmp, 4);

    // ͷ
    packetLSB = i2c_tmp[0];
    packetMSB = i2c_tmp[1];
    channelNumber = i2c_tmp[2];
    sequenceNumber = i2c_tmp[3];

    // 洢ͷϢ
    shtpHeader[0] = packetLSB;
    shtpHeader[1] = packetMSB;
    shtpHeader[2] = channelNumber;
    shtpHeader[3] = sequenceNumber;

    // ݳȣλ
    dataLength = ((unsigned int)packetMSB << 8) | packetLSB;
    dataLength &= ~(1 << 15); //  MSbitλ

    if (dataLength == 0)
    {
        return 0; // Ϊ
    }

    dataLength -= 4; // ȥͷ

    // ȡ
    getData(dataLength);

    return 1; // ɹݰ
}

// 򴫸ͶֱӴյֽ
// shtpData Ϊ MAX_PACKET_SIZEκֽڶʧ
// Arduino I2C ȡΪ 32 ֽڡͷ 4 ֽڣÿԶȡΪ 28 ֽ
char getData(unsigned int bytesRemaining)
{
	unsigned int dataSpot = 0;
    unsigned char incoming = 0;
    unsigned char x = 0;

    // ֿȡʣ
    while (bytesRemaining > 0)
    {
        unsigned int numberOfBytesToRead = bytesRemaining;

        // ÿζȡƣ4ֽڰͷ
        if (numberOfBytesToRead > (I2C_BUFFER_LENGTH - 4))
            numberOfBytesToRead = (I2C_BUFFER_LENGTH - 4);

        // ȡһݿ
        ReadNbyte(i2c_tmp, numberOfBytesToRead + 4);

        // ǰ4ֽڰͷȡЧ
        for (x = 0; x < numberOfBytesToRead; x++)
        {
            incoming = i2c_tmp[x + 4];
            if (dataSpot < MAX_PACKET_SIZE)
            {
                shtpData[dataSpot++] = incoming;
            }
        }

        bytesRemaining -= numberOfBytesToRead;
    }

    return 1; // ɹȡ
}

// ݰͱͷȻ
//  ACK򷵻 0
// TODO - Arduino Ϊ 32 ֽڡҪͷֽΪݰ
char sendPacket(unsigned char channelNumber, unsigned char dataLength)
{
	unsigned char packetLength,i;
    packetLength = dataLength + 4;

    // 4ֽڰͷ
    i2c_tmp[0] = (packetLength & 0xFF);             // LSB
    i2c_tmp[1] = 0;                                // MSB (̶Ϊ0)
    i2c_tmp[2] = channelNumber;                    // ͨ
    i2c_tmp[3] = (sequenceNumber[channelNumber]++); // кŵ

    // û
    for (i = 0; i < dataLength; i++)
    {
        i2c_tmp[4 + i] = shtpData[i];
    }

    // ݰ
    WriteNbyte(i2c_tmp, packetLength);

    return 1; // ɹ
}